"""Main module."""
# import dependencies
from pyomo.environ import ConcreteModel, Set, Var, Objective, minimize, maximize, Constraint
import numpy as np
import pandas as pd
from .constant import CET_ADDI, ORIENT_IO, ORIENT_OO, RTS_VRS, RTS_CRS, OPT_DEFAULT, OPT_LOCAL
from .utils import tools


class DEA:
    """Data Envelopment Analysis (DEA)
    """

    def __init__(self, y, x, orient, rts, yref=None, xref=None):
        """DEA: Envelopment problem

        Args:
            y (float): output variable.
            x (float): input variables.
            orient (String): ORIENT_IO (input orientation) or ORIENT_OO (output orientation)
            rts (String): RTS_VRS (variable returns to scale) or RTS_CRS (constant returns to scale)
            yref (String, optional): reference output. Defaults to None.
            xref (String, optional): reference inputs. Defaults to None.
        """
        # TODO(error/warning handling): Check the configuration of the model exist
        # Initialize DEA model
        self.orient = orient
        self.rts = rts
        self.y, self.x = tools.assert_valid_mupltiple_y_data(y, x)
        self.orient = orient
        self.rts = rts

        if type(yref) != type(None):
            self.yref, self.xref = tools.assert_valid_reference_data(self.y, self.x, yref, xref)
        else:
            self.yref, self.xref = self.y, self.x

  ## 当前被评价决策单元的序号 self.x[I0]
        # Initialize sets
        # print(self.xref)
        self.I = range(len(self.x))          ## I 是 被评价决策单元的数量
        self.__modeldict = {}
        for i in self.I:
            self.I0 = i

            self.__model__ = ConcreteModel()

            self.__model__.I2 = Set(initialize=range(len(self.xref)))      ## I2 是 参考决策单元的数量
            self.__model__.K = Set(initialize=range(len(self.x[0])))          ## K 是投入个数
            self.__model__.L = Set(initialize=range(len(self.y[0])))           ## L 是产出个数 被评价单元和参考单元的K，L一样

            # Initialize variable
            self.__model__.theta = Var(Set(initialize=range(1)),bounds=(None, None), doc='efficiency')
            self.__model__.lamda = Var(self.__model__.I2, bounds=(0.0, None), doc='intensity variables')

            # Setup the objective function and constraints
            if self.orient == ORIENT_IO:
                self.__model__.objective = Objective(
                    rule=self.__objective_rule(), sense=minimize, doc='objective function')
            else:
                self.__model__.objective = Objective(
                    rule=self.__objective_rule(), sense=maximize, doc='objective function')

            self.__model__.input = Constraint(
                self.__model__.K, rule=self.__input_rule(), doc='input constraint')
            self.__model__.output = Constraint(
                self.__model__.L, rule=self.__output_rule(), doc='output constraint')

            if self.rts == RTS_VRS:
                self.__model__.vrs = Constraint(
                     rule=self.__vrs_rule(), doc='various return to scale rule')

            self.__modeldict[i] = self.__model__

        # # Optimize model


    def __objective_rule(self):
        """Return the proper objective function"""

        def objective_rule(model):
            return model.theta[0]*1  + sum(model.lamda[i2] *0 for i2 in model.I2)

        return objective_rule


    def __input_rule(self):
        """Return the proper input constraint"""
        if self.orient == ORIENT_OO:
            def input_rule(model, k):
                return sum(model.lamda[i2] * self.xref[i2,k] for i2 in model.I2) <= self.x[self.I0,k]
        elif self.orient == ORIENT_IO:
            def input_rule(model, k):
                return sum(model.lamda[i2] * self.xref[i2,k] for i2 in model.I2) <= model.theta * self.x[self.I0,k]
        return input_rule



    def __output_rule(self):
        """Return the proper output constraint"""
        if self.orient == ORIENT_OO:
            def output_rule(model, l):
                return sum(model.lamda[i2] * self.yref[i2][l] for i2 in model.I2) >=model.theta * self.y[self.I0][l]
        elif self.orient == ORIENT_IO:
            def output_rule(model, l):
                return sum(model.lamda[i2] * self.yref[i2][l] for i2 in model.I2) >= self.y[self.I0][l]
        return output_rule

    def __vrs_rule(self):
        def vrs_rule(model):
            return sum(model.lamda[ i2] for i2 in model.I2) == 1

        return vrs_rule

    def optimize(self,  solver=OPT_DEFAULT):
        """Optimize the function by requested method

        Args:
            solver (string): The solver chosen for optimization. It will optimize with default solver if OPT_DEFAULT is given.
        """
        # TODO(error/warning handling): Check problem status after optimization
        
        problem_status,optimization_status,theta,lamda = {},{},{},{}
        for ind, problem in list(self.__modeldict.items()):
            problem_status[ind], optimization_status[ind] = tools.optimize_model(problem, CET_ADDI, solver)
            theta[ind],lamda[ind]=np.asarray(list(problem.theta[:].value)),np.asarray(list(problem.lamda[:].value))
        
        theta = pd.DataFrame(theta).T
        theta.columns = ["theta"]
        theta["eff"]=1/theta["theta"]
        lamda =pd.DataFrame(lamda) 
        lamda.columns = lamda.columns.map(lambda x : "Dum"+ str(x)+"'s lambda" ) 
        return problem_status,optimization_status,theta,lamda

    def info(self, dmu = "all"):
        """Show the infomation of the lp model

        Args:
            dmu (string): The solver chosen for optimization. Default is "all".
        """
        if dmu =="all":
            for ind, problem in list(self.__modeldict.items()):
                print(ind,"\n",problem.pprint())
 
        print(self.__modeldict[int(dmu)].pprint())




class RM(DEA):
    def __init__(self,  y, x,   orient, rts, yref=None, xref=None):
        """DEA: Directional distance function

        Args:
            y (float): output variable. 
            x (float): input variables.
            orient (String): ORIENT_IO (input orientation) or ORIENT_OO (output orientation)
            rts (String): RTS_VRS (variable returns to scale) or RTS_CRS (constant returns to scale)
            yref (String, optional): reference output. Defaults to None.
            xref (String, optional): reference inputs. Defaults to None.
        """
        # Initialize RM model

        self.orient = orient
        self.rts = rts
        self.y, self.x = tools.assert_valid_mupltiple_y_data(y, x)
        self.orient = orient
        self.rts = rts

        if type(yref) != type(None):
            self.yref, self.xref = tools.assert_valid_reference_data(self.y, self.x, yref, xref)
        else:
            self.yref, self.xref = self.y, self.x

  ## 当前被评价决策单元的序号 self.x[I0]
        # Initialize sets
        # print(self.xref)
        self.I = range(len(self.x))          ## I 是 被评价决策单元的数量
        self.__modeldict = {}
        for i in self.I:
            self.I0 = i

            self.__model__ = ConcreteModel()

            self.__model__.I2 = Set(initialize=range(len(self.xref)))      ## I2 是 参考决策单元的数量
            self.__model__.K = Set(initialize=range(len(self.x[0])))          ## K 是投入个数
            self.__model__.L = Set(initialize=range(len(self.y[0])))           ## L 是产出个数 被评价单元和参考单元的K，L一样

            # Initialize variable
            if self.orient == ORIENT_OO:
                self.__model__.theta = Var(self.__model__.L,bounds=(None, None), doc='efficiency')
            else:
                self.__model__.theta = Var(self.__model__.K,bounds=(None, None), doc='efficiency')
            
            self.__model__.lamda = Var(self.__model__.I2, bounds=(0.0, None), doc='intensity variables')

            # Setup the objective function  
            if self.orient == ORIENT_OO:
                self.__model__.objective = Objective(
                    rule=self.__objective_rule(), sense=maximize, doc='objective function')
            else:
                self.__model__.objective = Objective(
                    rule=self.__objective_rule(), sense=minimize, doc='objective function')
           
            # Setup the Constraint function  
            self.__model__.input = Constraint(
                self.__model__.K, rule=self.__input_rule(), doc='input constraint')
            self.__model__.output = Constraint(
                self.__model__.L, rule=self.__output_rule(), doc='output constraint')

            if self.rts == RTS_VRS:
                self.__model__.vrs = Constraint(
                     rule=self.__vrs_rule(), doc='various return to scale rule')

            self.__modeldict[i] = self.__model__

        # # Optimize model



    def __objective_rule(self):
        """Return the proper objective function"""
        if self.orient == ORIENT_OO:
            def objective_rule(model):
                return sum(model.theta[l]*1 for l in model.L) / len(model.L)
        else:
            def objective_rule(model):
                return sum(model.theta[k]*1 for k in model.K) / len(model.K)
        return objective_rule


    def __input_rule(self):
        """Return the proper input constraint"""
        if self.orient == ORIENT_OO:
            def input_rule(model, k):
                return sum(model.lamda[i2] * self.xref[i2,k] for i2 in model.I2) <= self.x[self.I0,k]
        elif self.orient == ORIENT_IO:
            def input_rule(model, k):
                return sum(model.lamda[i2] * self.xref[i2,k] for i2 in model.I2) <= model.theta[k] * self.x[self.I0,k]
        return input_rule

    def __output_rule(self):
        """Return the proper output constraint"""
        if self.orient == ORIENT_OO:
            def output_rule(model, l):
                return sum(model.lamda[i2] * self.yref[i2][l] for i2 in model.I2) >=model.theta[l] * self.y[self.I0][l]
        elif self.orient == ORIENT_IO:
            def output_rule(model, l):
                return sum(model.lamda[i2] * self.yref[i2][l] for i2 in model.I2) >= self.y[self.I0][l]
        return output_rule


    def __vrs_rule(self):
        def vrs_rule(model):
            return sum(model.lamda[ i2] for i2 in model.I2) == 1

        return vrs_rule


    def optimize(self,  solver=OPT_DEFAULT):
        """Optimize the function by requested method

        Args:
            solver (string): The solver chosen for optimization. It will optimize with default solver if OPT_DEFAULT is given.
        """
        # TODO(error/warning handling): Check problem status after optimization
        
        problem_status,optimization_status,theta,lamda = {},{},{},{}
        for ind, problem in list(self.__modeldict.items()):
            problem_status[ind], optimization_status[ind] = tools.optimize_model(problem, CET_ADDI, solver)
            theta[ind],lamda[ind]=np.asarray(list(problem.theta[:].value)),np.asarray(list(problem.lamda[:].value))
        
        theta = pd.DataFrame(theta).T
        theta.columns = theta.columns.map(lambda x : "Input"+ str(x)+"'s theta" ) 
        for index,col in enumerate( theta.columns):
            theta["Input{}'s eff".format(index)]=theta[col]

        lamda =pd.DataFrame(lamda) 
        lamda.columns = lamda.columns.map(lambda x : "Dum"+ str(x)+"'s lambda" ) 
        return problem_status,optimization_status,theta,lamda

    def info(self, dmu = "all"):
        """Show the infomation of the lp model

        Args:
            dmu (string): The solver chosen for optimization. Default is "all".
        """
        if dmu =="all":
            for ind, problem in list(self.__modeldict.items()):
                print(ind,"\n",problem.pprint())
 
        print(self.__modeldict[int(dmu)].pprint())





class DDF(DEA):
    def __init__(self,  y, x, b=None, gy=[1], gx=[1], gb=None, rts=RTS_VRS, yref=None, xref=None, bref=None):
        """DEA: Directional distance function

        Args:
            y (float): output variable. 
            x (float): input variables.
            b (float), optional): undesirable output variables. Defaults to None.
            gy (list, optional): output directional vector. Defaults to [1].
            gx (list, optional): input directional vector. Defaults to [1].
            gb (list, optional): undesirable output directional vector. Defaults to None.
            rts (String): RTS_VRS (variable returns to scale) or RTS_CRS (constant returns to scale)
            yref (String, optional): reference output. Defaults to None.
            xref (String, optional): reference inputs. Defaults to None.
            bref (String, optional): reference undesirable output. Defaults to None.
        """
        # Initialize DEA model
        self.__model__ = ConcreteModel()

        self.y, self.x, self.b, self.gy, self.gx, self.gb = tools.assert_valid_direciontal_data(
            y, x, b, gy, gx, gb)
        self.rts = rts

        if type(yref) != type(None):
            self.yref, self.xref, self.bref = tools.assert_valid_reference_data_with_bad_outputs(
                self.y, self.x, self.b, yref, xref, bref)
        else:
            self.yref, self.xref, self.bref = self.y, self.x, self.b
        self.__model__.R = Set(initialize=range(len(self.yref)))

        # Initialize sets
        self.__model__.I = Set(initialize=range(len(self.y)))
        self.__model__.J = Set(initialize=range(len(self.x[0])))
        self.__model__.K = Set(initialize=range(len(self.y[0])))
        if type(b) != type(None):
            self.__model__.L = Set(initialize=range(len(self.b[0])))

        # Initialize variable
        self.__model__.theta = Var(
            self.__model__.I, doc='directional distance')

        self.__model__.lamda = Var(self.__model__.I, self.__model__.R, bounds=(
            0.0, None), doc='intensity variables')

        # Setup the objective function and constraints
        self.__model__.objective = Objective(
            rule=self._DEA__objective_rule(), sense=maximize, doc='objective function')
        self.__model__.input = Constraint(
            self.__model__.I, self.__model__.J, rule=self.__input_rule(), doc='input constraint')
        self.__model__.output = Constraint(
            self.__model__.I, self.__model__.K, rule=self.__output_rule(), doc='output constraint')

        if type(b) != type(None):
            self.__model__.undesirable_output = Constraint(
                self.__model__.I, self.__model__.L, rule=self.__undesirable_output_rule(), doc='undesirable output constraint')

        if self.rts == RTS_VRS:
            self.__model__.vrs = Constraint(
                self.__model__.I, rule=self.__vrs_rule(), doc='various return to scale rule')

        # Optimize model
        self.optimization_status = 0
        self.problem_status = 0

    def __input_rule(self):
        """Return the proper input constraint"""
        def input_rule(model, o, j):
            return self.x[o][j] - model.theta[o]*self.gx[j] >= sum(model.lamda[o, r] * self.xref[r][j] for r in model.R)
        return input_rule

    def __output_rule(self):
        """Return the proper output constraint"""
        def output_rule(model, o, k):
            return self.y[o][k] + model.theta[o]*self.gy[k] <= sum(model.lamda[o, r] * self.yref[r][k] for r in model.R)
        return output_rule

    def __undesirable_output_rule(self):
        """Return the proper undesirable output constraint"""
        def undesirable_output_rule(model, o, l):
            return self.b[o][l] - model.theta[o]*self.gb[l] == sum(model.lamda[o, r] * self.bref[r][l] for r in model.R)
        return undesirable_output_rule

    def __vrs_rule(self):
        """Return the VRS constraint"""
        def vrs_rule(model, o):
            return sum(model.lamda[o, r] for r in model.R) == 1
        return vrs_rule




